unit CityHash;

interface

{.$DEFINE _C_NAME}

{$DEFINE CITY_HASH_FROM_DLL}
{.$DEFINE CITY_HASH_FROM_OBJ}
{.$DEFINE CITY_HASH_FROM_PAS}

{$IFNDEF CITY_HASH_FROM_PAS}
  {$IFNDEF CITY_HASH_FROM_OBJ}
    {$DEFINE CITY_HASH_FROM_DLL}
  {$ENDIF}
{$ENDIF}

{$IFDEF CITY_HASH_FROM_DLL}
  {$UNDEF CITY_HASH_FROM_OBJ}
  {$UNDEF CITY_HASH_FROM_PAS}
{$ELSE}
  {$IFDEF CITY_HASH_FROM_OBJ}
    {$UNDEF CITY_HASH_FROM_DLL}
    {$UNDEF CITY_HASH_FROM_PAS}
  {$ELSE}
    {$UNDEF CITY_HASH_FROM_DLL}
    {$UNDEF CITY_HASH_FROM_OBJ}
  {$ENDIF}
{$ENDIF}

type
  size_t  = System.Cardinal;
  uint8   = Byte;
  uint16  = Word;
  uint32  = Cardinal;
  uint64  = System.UInt64;
  uint128 = record first: uint64; second: uint64; end;
  p_uint128 = ^uint128;

{$IFDEF CITY_HASH_FROM_DLL}  
const
  CityHashLib = {$IFDEF _C_NAME} 'CityHash_.dll' {$ELSE} 'CityHash.dll' {$ENDIF};

function CityHash32(const buf: pointer; len: size_t): uint32; cdecl; external CityHashLib {$IFDEF _C_NAME} name '_CityHash32' {$ENDIF};
function CityHash64(const buf: pointer; len: size_t): uint64; cdecl; external CityHashLib {$IFDEF _C_NAME} name '_CityHash64' {$ENDIF};
function CityHash64WithSeed(const buf: pointer; len: size_t; seed: uint64): uint64; cdecl; external CityHashLib {$IFDEF _C_NAME} name '_CityHash64WithSeed' {$ENDIF};
function CityHash64WithSeeds(const buf: pointer; len: size_t; seed0, seed1: uint64): uint64; cdecl; external CityHashLib {$IFDEF _C_NAME} name '_CityHash64WithSeeds' {$ENDIF};
function CityHash128(const buf: pointer; len: size_t): uint128; cdecl; external CityHashLib {$IFDEF _C_NAME} name '_CityHash128' {$ENDIF};
function CityHash128WithSeed(const buf: pointer; len: size_t; seed: uint128): uint128; cdecl; external CityHashLib {$IFDEF _C_NAME} name '_CityHash128WithSeed' {$ENDIF};
{$ELSE}
function CityHash32(const buf: pointer; len: size_t): uint32; {$IFDEF CITY_HASH_FROM_OBJ} cdecl; external; {$ENDIF}
function CityHash64(const buf: pointer; len: size_t): uint64; {$IFDEF CITY_HASH_FROM_OBJ} cdecl; external; {$ENDIF}
function CityHash64WithSeed(const buf: pointer; len: size_t; seed: uint64): uint64; {$IFDEF CITY_HASH_FROM_OBJ} cdecl; external; {$ENDIF}
function CityHash64WithSeeds(const buf: pointer; len: size_t; seed0, seed1: uint64): uint64; {$IFDEF CITY_HASH_FROM_OBJ} cdecl; external; {$ENDIF}
function CityHash128(const buf: pointer; len: size_t): uint128; {$IFDEF CITY_HASH_FROM_OBJ} cdecl; external; {$ENDIF}
function CityHash128WithSeed(const buf: pointer; len: size_t; seed: uint128): uint128; {$IFDEF CITY_HASH_FROM_OBJ} cdecl; external; {$ENDIF}
{$ENDIF}

function Uint128Low64(const x: p_uint128): uint64; inline;
function Uint128High64(const x: p_uint128): uint64; inline;

function Hash128to64(const x: p_uint128): uint64; inline;

implementation

const
  kMul = $9ddfea08eb382d69;

function Uint128Low64(const x: p_uint128): uint64; inline;
begin
  Result := x.first;
end;

function Uint128High64(const x: p_uint128): uint64; inline;
begin
  Result := x.second;
end;

function Hash128to64(const x: p_uint128): uint64; inline;
var
  a, b: uint64;
begin
  // Murmur-inspired hashing.
  a := (x.first xor x.second) * uint64(kMul);
  a := a xor (a shr 47);
  b := (x.second xor a) * uint64(kMul);
  b := b xor (b shr 47);
  Result := b * uint64(kMul);
end;

{$IFDEF CITY_HASH_FROM_OBJ}

{$LINK win32\city.obj}

procedure _llmul;
asm
  jmp System.@_llmul
end;

{$ENDIF}

{$IFDEF CITY_HASH_FROM_PAS}

procedure swap(var a, b: uint64); inline;
var
  x: uint64;
begin
  x := a;
  a := b;
  b := x;
end;

function bswap_32(i: uint32): uint32; assembler; register;
asm
  bswap eax
end;

function bswap_64(Value: uint64): uint64; assembler; register;
asm
  mov edx, [esp+8]
  bswap edx
  mov eax, [esp+12]
  bswap eax
end;

//==============================================================================

function HashLen16(const u, v, mul: uint64): uint64; inline; overload;
var
  a, b : uint64;
begin
  // Murmur-inspired hashing.
  a := (u xor v) * mul;
  a := a xor (a shr 47);
  b := (v xor a) * mul;
  b := b xor (b shr 47);
  Result := b * mul;
end;

function HashLen16(const L, H: uint64): uint64; inline; overload;
var
  a, b: uint64;
begin
  a := (L xor H) * kMul;
  a := a xor (a shr 47);
  b := (H xor a) * kMul;
  b := b xor (b shr 47);
  Result := b * kMul;
end;

// Some primes between 2^63 and 2^64 for various uses.
const
  k0 = $c3a5c85c97cb3127;
  k1 = $b492b66fbe98f273;
  k2 = $9ae16a3b2f90404f;

// Bitwise right rotate.  Normally this will compile to a single
// instruction, especially if the shift is a manifest constant.
function Rotate(const val: uint64; const shift: Integer): uint64; inline;
begin
  // Avoid shifting by 64: doing so yields an undefined result.
  if shift = 0 then begin
    Result := val;
  end else begin
    Result := (val shr shift) or (val shl (64 - shift));
  end;
end;

function ShiftMix(const val: uint64): uint64; inline;
begin
  Result := val xor (val shr 47);
end;

type
  a4 = packed record
    s: array[0..3] of Byte;
  end;

function HashLen0to16(const s: pointer; const len: size_t): uint64; inline;
var
  mul,a,b,c,d,Y,Z: uint64;
begin
  if (len >= 8) then begin
    mul := k2 + len * 2;
    a := uint64(s^) + k2;
    b := uint64(Ptr(uint32(s) + len - 8)^);
    c := Rotate(b, 37) * mul + a;
    d := (Rotate(a, 25) + b) * mul;
    Result:= HashLen16(c, d, mul);
  end else if (len >= 4) then begin
    mul := k2 + len * 2;
    a := uint32(s^);
    Result:= HashLen16(len + (a shl 3), uint32(Ptr(uint32(s) + len - 4)^), mul);
  end else if (len > 0) then begin
    {uint8}  a := a4(s^).s[0];         //s[0];
    {uint8}  b := a4(s^).s[len shr 1]; //s[len shr 1];
    {uint8}  c := a4(s^).s[len - 1];   //s[len - 1];
    {uint32} y := (a) + ((b) shl 8);
    {uint32} z := len + ((c) shl 2);
    Result:= ShiftMix(y * k2 xor z * k0) * k2;
  end else begin
    Result := uint64(k2);
  end;
end;

// This probably works well for 16-byte strings as well, but it may be overkill
// in that case.
function HashLen17to32(const s: pointer; const len: size_t): uint64; inline;
var
  mul,a,b,c,d: uint64;
begin
  mul := k2 + len * 2;
  a := uint64(s^) * k1;
  b := uint64(Ptr(uint32(s) + 8)^);
  c := uint64(Ptr(uint32(s) + len -  8)^) * mul;
  d := uint64(Ptr(uint32(s) + len - 16)^) * k2;
  Result :=
    HashLen16(
      Rotate(a + b, 43) + Rotate(c, 30) + d,
      a + Rotate(b + k2, 18) + c,
      mul
    );
end;

// Return an 8-byte hash for 33 to 64 bytes.
function HashLen33to64(const s: pointer; const len: size_t): uint64; inline;
var
  mul,a,b,c,d,e,f,g,h,u,v,w,x,y,z: uint64;
begin
  {uint64} mul := k2 + len * 2;
  {uint64} a := uint64(s^) * k2;
  {uint64} b := uint64(Ptr(uint32(s) + 8)^);
  {uint64} c := uint64(Ptr(uint32(s) + len - 24)^);
  {uint64} d := uint64(Ptr(uint32(s) + len - 32)^);
  {uint64} e := uint64(Ptr(uint32(s) + 16)^) * k2;
  {uint64} f := uint64(Ptr(uint32(s) + 24)^) * 9;
  {uint64} g := uint64(Ptr(uint32(s) + len - 8)^);
  {uint64} h := uint64(Ptr(uint32(s) + len - 16)^) * mul;

  {uint64} u := Rotate(a + g, 43) + (Rotate(b, 30) + c) * 9;
  {uint64} v := ((a + g) xor d) + f + 1;
  {uint64} w := bswap_64((u + v) * mul) + h;
  {uint64} x := Rotate(e + f, 42) + c;
  {uint64} y := (bswap_64((v + w) * mul) + g) * mul;
  {uint64} z := e + f + c;

  a := bswap_64((x + z) * mul + y) + b;
  b := ShiftMix((z + a) * mul + d + h) * mul;
  Result := b + x;
 end;

// Return a 16-byte hash for 48 bytes.  Quick and dirty.
// Callers do best to use "random-looking" values for a and b.
procedure WeakHashLen32WithSeeds(const w, x, y, z: uint64; a, b: uint64; var res: uint128); inline; overload;
var
  c : uint64;
begin
  a := a + w;
  b := Rotate(b + a + z, 21);
  c := a;
  a := a + x + y;
  b := b + Rotate(a, 44);
  res.first := a + z;
  res.second:= b + c;
end;

// Return a 16-byte hash for s[0] ... s[31], a, and b.  Quick and dirty.
procedure WeakHashLen32WithSeeds(const s: pointer; const a, b: uint64; var res: uint128); inline; overload;
begin
    WeakHashLen32WithSeeds(
      uint64(s^),
      uint64(Ptr(uint32(s) + 8)^),
      uint64(Ptr(uint32(s) + 16)^),
      uint64(Ptr(uint32(s) + 24)^),
      a,
      b,
      res
    );
 end;

function CityHash64(const buf: pointer; len: size_t): uint64;
var
  x, y, z: uint64;
  v, w: uint128;
  s: pointer;
begin
  s := buf;
  
  if len <= 32 then begin
    if len <= 16 then begin
      Result := HashLen0to16(s, len);
      Exit;
    end else begin
      Result := HashLen17to32(s, len);
      Exit;
    end;
  end else if len <= 64 then begin
    Result := HashLen33to64(s, len);
    Exit;
  end;

  // For strings over 64 bytes we hash the end first, and then as we
  // loop we keep 56 bytes of state: v, w, x, y, and z.
  x := uint64(Ptr(uint32(s) + len - 40)^);
  y := uint64(Ptr(uint32(s) + len - 16)^) + uint64(Ptr(uint32(s) + len - 56)^);
  z := HashLen16(uint64(Ptr(uint32(s) + len - 48)^) + len, uint64(Ptr(uint32(s) + len - 24)^));
  WeakHashLen32WithSeeds(Ptr(uint32(s) + len - 64), len, z, v);
  WeakHashLen32WithSeeds(Ptr(uint32(s) + len - 32), y + k1, x, w);
  x := x * k1 + uint64(s^);

  // Decrease len to the nearest multiple of 64, and operate on 64-byte chunks.
  len := (len - 1) and not $0000003f;

  while len > 0 do begin
    x := Rotate(x + y + v.first  + uint64(Ptr(uint32(s) +  8)^), 37) * k1;
    y := Rotate(y + v.second + uint64(Ptr(uint32(s) + 48)^), 42) * k1;
    x := x xor w.second;
    y := y + v.first  + uint64(Ptr(uint32(s) + 40)^);
    z := Rotate(z + w.first, 33) * k1;
    WeakHashLen32WithSeeds(s, v.second * k1, x + w.first, v);
    WeakHashLen32WithSeeds(Ptr(uint32(s) + 32), z + w.second, y + uint64(Ptr(uint32(s) + 16)^), w);
    swap(z, x);
    s := Ptr(uint32(s) + 64);
    len := len - 64;
  end;

  Result :=
    HashLen16(
      HashLen16(v.first,  w.first)  + ShiftMix(y) * k1 + z,
      HashLen16(v.second, w.second) + x
    );
end;

function CityHash64WithSeed(const buf: pointer; len: size_t; seed: uint64): uint64;
begin
  Result := HashLen16(CityHash64(buf, len) - k2, seed);
end;

function CityHash64WithSeeds(const buf: pointer; len: size_t; seed0, seed1: uint64): uint64;
begin
  Result := HashLen16(CityHash64(buf, len) - seed0, seed1);
end;

//==============================================================================

// A subroutine for CityHash128().  Returns a decent 128-bit hash for strings
// of any length representable in signed long.  Based on City and Murmur.
function CityMurmur( s: pointer;  len:size_t;  seed:uint128):uint128;
var
  a,b,c,d : uint64;
  l:LongInt;
begin
  {uint64} a := seed.first;//Uint128Low64(@seed);
  {uint64} b := seed.second;//Uint128High64(@seed);
  {LongInt}l := len - 16;

  if (len <= 16) then
  begin
    a := ShiftMix(a * k1) * k1;
    c := b * k1 + HashLen0to16(s, len);
    //d := ShiftMix(a + (len >= 8 ? Fetch64(s) : c));
    if len >= 8
    then d := ShiftMix(a + uint64(s^))
    else d := ShiftMix(a + c);
  end
  else
  begin// len > 16
    c := HashLen16(uint64(Ptr(uint32(s) + len - 8)^) + k1, a);
    d := HashLen16(b + len, c + uint64(Ptr(uint32(s) + len - 16)^));
    a:= a + d;

    repeat
      a:= (a xor ShiftMix(uint64(s^) * k1) * k1){;a:= a} * k1;
      b:= b xor a;
      c:= (c xor ShiftMix(uint64(Ptr(uint32(s) + 8)^) * k1) * k1){; c:= c} * k1;
      d:= d xor c;
      s:= Ptr(uint32(s) + 16);
      l:= l - 16;
    until l<=0;//while (l > 0);
  end;
  a := HashLen16(a, c);
  b := HashLen16(d, b);

//result:= uint128(a xor b, HashLen16(b, a));
  result.first  := a xor b;
  result.second := HashLen16(b, a);
 end;

function CityHash128WithSeed(const buf: pointer; len: size_t; seed: uint128): uint128;
var
  v, w :uint128;
  X,Y,Z : uint64;
  ZX:uint64;
  tail_done : size_t;
  s: Pointer;
begin
  s := buf;

  if (len < 128) then  begin
    result:= CityMurmur(s, len, seed); exit;
  end;

  // We expect len >= 128 to be the common case.  Keep 56 bytes of state:
  // v, w, x, y, and z.
//  pair<uint64, uint64> v, w;

  {uint64} x := seed.first;//Uint128Low64(@seed);
  {uint64} y := seed.second;//Uint128High64(@seed);
  {uint64} z := len * k1;

  v.first := Rotate(y xor k1, 49) * k1 + uint64(s^);
  v.second:= Rotate(v.first, 42) * k1 + uint64(Ptr(uint32(s) + 8)^);
  w.first := Rotate(y + z, 35) * k1 + x;
  w.second:= Rotate(x + uint64(Ptr(uint32(s) + 88)^), 53) * k1;

  // This is the same inner loop as CityHash64(), manually unrolled.
  while (len >= 128)
  do begin
    {z}zx:=  k1*Rotate(z + w.first, 33);
    {x}z := (k1*Rotate(x + y + v.first + uint64(Ptr(uint32(s) + 8)^), 37)){;x:= x} xor w.second;
    x:=zx;

    y :=  k1*Rotate(y + v.second + uint64(Ptr(uint32(s) + 48)^), 42){;
    y:= y}            + v.first  + uint64(Ptr(uint32(s) + 40)^);

    {v := }WeakHashLen32WithSeeds(s, v.second * k1,    {x}z + w.first, v);
    {w := }WeakHashLen32WithSeeds(Ptr(uint32(s) + 32), {z}x + w.second, y + uint64(Ptr(uint32(s)+ 16)^), w);

  //zx:=z;z:=x;x:=zx;//std.swap(z, x);

    s:= Ptr(uint32(s)+ 64);

    {x}zx := Rotate(x + y + v.first + uint64(Ptr(uint32(s) +  8)^), 37) * k1{;    x:= x} xor w.second;
    y := Rotate(    y + v.second+ uint64(Ptr(uint32(s) + 48)^), 42) * k1{;
    y:= y} +            v.first + uint64(Ptr(uint32(s) + 40)^);

    {z}x := Rotate(z + w.first, 33) * k1;
    z:=zx; {zx:=x;x:=z;z:=zx;}//std.swap(z, x);

    {v := }WeakHashLen32WithSeeds(s,          v.second * k1, {x}z + w.first, v);
    {w := }WeakHashLen32WithSeeds(Ptr(uint32(s) + 32),       {z}x + w.second, y + uint64(Ptr(uint32(s) + 16)^), w);

    s:= Ptr(uint32(s) + 64);
    len:= len - 128;
  end;

  x :=      x + Rotate(v.first + z, 49) * k0;
  y := y * k0 + Rotate(w.second, 37);
  z := z * k0 + Rotate(w.first, 27);

  w.first:= w.first * 9;
  v.first:= v.first * k0;
  // If 0 < len < 128, hash up to 4 chunks of 32 bytes each from the end of s.

//for (size_t tail_done = 0; tail_done < len; ) {
//tail_done += 32;
  tail_done:=0;
  while tail_done < len do
  begin
    tail_done:= tail_done + 32;
    y := Rotate(x + y, 42) * k0 + v.second;
    w.first:= w.first + uint64(Ptr(uint32(s) + len - tail_done + 16)^);
    x := x * k0 + w.first;
    z := z + w.second + uint64(Ptr(uint32(s) + len - tail_done)^);
    w.second:= w.second + v.first;

    {v := }WeakHashLen32WithSeeds(Ptr(uint32(s) + len - tail_done), v.first + z, v.second, v);
    v.first:= v.first * k0;
  end;

  // At this point our 56 bytes of state should contain more than
  // enough information for a strong 128-bit hash.  We use two
  // different 56-byte-to-8-byte hashes to get a 16-byte final result.
  x := HashLen16(x,     v.first);
  y := HashLen16(y + z, w.first);
//result:= uint128(HashLen16(x + v.second, w.second) + y,
//                 HashLen16(x + w.second, y + v.second));
  result.first := HashLen16(x + v.second, w.second) + y;
  result.second:= HashLen16(x + w.second, v.second  + y );
end;

function CityHash128(const buf: pointer; len: size_t): uint128;
var v_uint128 : uint128;
BEGIN
  IF len >= 16
  then begin
    v_uint128.first  :=uint64(buf^);
    v_uint128.second :=uint64(Ptr(uint32(buf)+8)^)+k0;
    result:=CityHash128WithSeed(Ptr(uint32(buf)+16),len-16, v_uint128);
  end
  else begin
    v_uint128.first  :=uint64(k0);
    v_uint128.second :=uint64(k1);
    result:=CityHash128WithSeed(buf,   len,    v_uint128);
  end;
end;

const
// Magic numbers for 32-bit hashing.  Copied from Murmur3.
{uint32_t} c1 = $cc9e2d51;
{uint32_t} c2 = $1b873593;

// A 32-bit to 32-bit integer hash copied from Murmur3.
function fmix(h: uint32): uint32; inline;//
begin
{
  h:= h xor h shr 16;
  h:= h * $85ebca6b;
  h:= h xor h shr 13;
  h:= h * $c2b2ae35;
  h:= h xor h shr 16;
  result:= h;
}
  h:=     (h xor h shr 16) * $85ebca6b;
  h:=     (h xor h shr 13) * $c2b2ae35;
  result:= h xor h shr 16;
end;

function Rotate32(const val: uint32; shift: Integer): uint32;  inline;// //ok
begin
{  if shift = 0
  then result:=val
  else} result:=(val shr shift) or (val shl (32 - shift));
end;

function Mur(a: uint32; h: uint32): uint32; inline;//
begin
  // Helper from Murmur3 for combining two 32-bit values.
{
  a := a * c1;
  a := Rotate32(a, 17);
  a := a * c2;
  h := h xor a;
  h := Rotate32(h, 19);
  result:= h * 5 + $e6546b64;
}
  result:= Rotate32(h xor (Rotate32(a * c1, 17) * c2), 19) * 5 + $e6546b64;
end;

type
  tab=packed record
    s :array[0..4] of ShortInt;
  end;

function Hash32Len0to4(s:pointer; len: size_t): uint32; inline;//
var
  b,c : uint32;
  i : Integer;
//v : ShortInt;
begin
  {uint32} b := 0;
  {uint32} c := 9;
  for i := 0 to len-1 do
  begin
    b := b * c1 + tab(s^).s[i];//v;
    c := c xor b;
  end;
  result:= fmix(Mur(b, Mur(len, c)));
end;

function Hash32Len5to12(s:pointer; len: size_t): uint32; inline;//
var
  d: uint32;
begin
(***
  {uint32}a := len; b := len * 5; c := 9; d := b;
  a:= a + {Fetch32}uint32(s^);
  b:= b + {Fetch32}uint32(Ptr(uint32(s) + len - 4)^);
  c:= c + {Fetch32}uint32(Ptr(uint32(s) + ((len shr 1) and 4))^);

  result:= fmix(Mur(c, Mur(b, Mur(a, d))));
**)
  d := len * 5;

  result:= fmix(
  Mur({c,}9  +uint32(Ptr(uint32(s) + ((len shr 1) and 4))^),
  Mur({b,}d  +uint32(Ptr(uint32(s) + len - 4)^),
  Mur({a,}len+uint32(s^),
      d
  ))));
end;

function Hash32Len13to24(s:pointer; len: size_t): uint32; inline;//
//var
//   a,b,c,d,e,f,h : uint32;
begin
(**
  {uint32} a := {Fetch32}uint32(Ptr(uint32(s) - 4 + (len shr 1))^);
  {uint32} b := {Fetch32}uint32(Ptr(uint32(s) + 4)^);
  {uint32} c := {Fetch32}uint32(Ptr(uint32(s) + len - 8)^);
  {uint32} d := {Fetch32}uint32(Ptr(uint32(s) + (len shr 1))^);
  {uint32} e := {Fetch32}uint32(s^);
  {uint32} f := {Fetch32}uint32(Ptr(uint32(s) + len - 4)^);
  {uint32} h := len;

  result:= fmix(Mur(f, Mur(e, Mur(d, Mur(c, Mur(b, Mur(a, h)))))));
***)
  result:=
  fmix(
  Mur({f,}uint32(Ptr(uint32(s) + len - 4)^),
  Mur({e,}uint32(s^),
  Mur({d,}uint32(Ptr(uint32(s) + (len shr 1))^),
  Mur({c,}uint32(Ptr(uint32(s) + len - 8)^),
  Mur({b,}uint32(Ptr(uint32(s) + 4)^),
  Mur({a,}uint32(Ptr(uint32(s) - 4 + (len shr 1))^),
      {h }len
  )))))));
end;

function CityHash32(const buf: pointer; len: size_t): uint32;
var
  h,g,f, a0,a1,a2,a3,a4 : uint32;
  iters : size_t;
  ab : uint32;
  b,c,d : uint32;
  i : Integer;
  s: Pointer;
begin
  s := buf;
  if (len <= 24) then  begin
{
     if (len <= 12) then
     begin
       if (len <= 4)
       then result:=Hash32Len0to4(s, len)
       else result:=Hash32Len5to12(s, len)
     end
     else
     result:=Hash32Len13to24(s, len);
     exit;
}
     if (len <= 12) then
     begin
       if (len <= 4)
       then begin
//         result:=Hash32Len0to4(s, len);
    b := 0;
    c := 9;
    for i := 0 to len-1 do
    begin
      b := b * c1 + tab(s^).s[i];//v;
      c := c xor b;
    end;
    result:= fmix(mur(b, mur(len, c)));

         exit;
       end
       else begin
//         result:=Hash32Len5to12(s, len);
  d := len * 5;
  result:= fmix(
  Mur({c,}9  +uint32(Ptr(uint32(s) + ((len shr 1) and 4))^),
  Mur({b,}d  +uint32(Ptr(uint32(s) + len - 4)^),
  Mur({a,}len+uint32(s^),
      d
  ))));
         exit;
       end;
     end
     else begin
//       result:=Hash32Len13to24(s, len);
  result:=
  fmix(
  Mur({f,}uint32(Ptr(uint32(s) + len - 4)^),
  Mur({e,}uint32(s^),
  Mur({d,}uint32(Ptr(uint32(s) + (len shr 1))^),
  Mur({c,}uint32(Ptr(uint32(s) + len - 8)^),
  Mur({b,}uint32(Ptr(uint32(s) + 4)^),
  Mur({a,}uint32(Ptr(uint32(s) - 4 + (len shr 1))^),
      {h }len
  )))))));
       exit;
     end;
  end;

  // len > 24
  {uint32} h := len; g := c1 * len; f := g;
  {uint32} a0 := Rotate32({Fetch32}uint32(Ptr(uint32(s) + len -  4)^) * c1, 17) * c2;
  {uint32} a1 := Rotate32({Fetch32}uint32(Ptr(uint32(s) + len -  8)^) * c1, 17) * c2;
  {uint32} a2 := Rotate32({Fetch32}uint32(Ptr(uint32(s) + len - 16)^) * c1, 17) * c2;
  {uint32} a3 := Rotate32({Fetch32}uint32(Ptr(uint32(s) + len - 12)^) * c1, 17) * c2;
  {uint32} a4 := Rotate32({Fetch32}uint32(Ptr(uint32(s) + len - 20)^) * c1, 17) * c2;
{
  h := h xor a0;
  h := Rotate32(h, 19);
  h := h * 5 + $e6546b64;
  h := h xor a2;
  h := Rotate32(h, 19);
  h := h * 5 + $e6546b64;

  g := g xor a1;
  g := Rotate32(g, 19);
  g := g * 5 + $e6546b64;
  g := g xor a3;
  g := Rotate32(g, 19);
  g := g * 5 + $e6546b64;

  f := f + a4;
  f := Rotate32(f, 19);
  f := f * 5 + $e6546b64;
}
  h := Rotate32((Rotate32(h xor a0, 19) * 5 + $e6546b64) xor a2, 19) * 5 + $e6546b64;
  g := Rotate32((Rotate32(g xor a1, 19) * 5 + $e6546b64) xor a3, 19) * 5 + $e6546b64;
  f := Rotate32(f + a4, 19) * 5 + $e6546b64;

  iters := (len - 1) div 20;
  repeat
    a0 := Rotate32(           uint32(s^) * c1, 17) * c2;
    a1 :=          uint32(Ptr(uint32(s) +  4)^);
    a2 := Rotate32(uint32(Ptr(uint32(s) +  8)^) * c1, 17) * c2;
    a3 := Rotate32(uint32(Ptr(uint32(s) + 12)^) * c1, 17) * c2;
    a4 :=          uint32(Ptr(uint32(s) + 16)^);
{
    h := h xor a0;
    h := Rotate32(h, 18);
    h := h * 5 + $e6546b64;

    f := f + a1;
    f := Rotate32(f, 19);
    f := f * c1;

    g := g + a2;
    g := Rotate32(g, 18);
    g := g * 5 + $e6546b64;

    h := h xor (a3 + a1); //h := h xor a3 + a1;  error
    h := Rotate32(h, 19);
    h := h * 5 + $e6546b64;

    g := g xor a4;
    g := bswap_32(g) * 5;

    h := h + a4 * 5;
    h := bswap_32(h);

    f := f + a0;
 }
 { 
    h := h xor a0;
    h := Rotate32(h, 18);
    h := h * 5 + $e6546b64;
    h := h xor (a3 + a1); //h := h xor a3 + a1;  error
    h := Rotate32(h, 19);
    h := h * 5 + $e6546b64;
    h := h + a4 * 5;
    h := bswap_32(h);

    f := f + a1;
    f := Rotate32(f, 19);
    f := f * c1;
    f := f + a0;

    g := g + a2;
    g := Rotate32(g, 18);
    g := g * 5 + $e6546b64;
    g := g xor a4;
    g := bswap_32(g) * 5;
}
    {h}ab := bswap_32(Rotate32((Rotate32(h xor a0, 18) * 5 + $e6546b64) xor (a3 + a1), 19 ) * 5 + $e6546b64 + a4 * 5);
    {f}h  := Rotate32(f + a1, 19) * c1 + a0;
    {g}f  := bswap_32((Rotate32(g + a2, 18) * 5 + $e6546b64) xor a4) * 5;
    g:=ab;
//  PERMUTE3(f, h, g);  ab:=h; h:=f;  f:=g;  g:=ab;

    s:= Ptr(uint32(s) + 20);
    iters := iters -1;
  until  iters <= 0; //while (--iters <> 0);
{
  g := Rotate32(g, 11) * c1;
  g := Rotate32(g, 17) * c1;
  f := Rotate32(f, 11) * c1;
  f := Rotate32(f, 17) * c1;

  h := Rotate32(h + g, 19);
  h := h * 5 + $e6546b64;
  h := Rotate32(h, 17) * c1;
  h := Rotate32(h + f, 19);
  h := h * 5 + $e6546b64;
  h := Rotate32(h, 17) * c1;
  result:= h;
}
{ g f h
  g := Rotate32(Rotate32(g, 11) * c1, 17) * c1;
  f := Rotate32(Rotate32(f, 11) * c1, 17) * c1;

  result:=
  Rotate32(
  Rotate32(
  Rotate32(
  Rotate32(h + g, 19) * 5
  + $e6546b64, 17) * c1 + f, 19) * 5
  + $e6546b64, 17) * c1;
}
  result:=
  Rotate32(
  Rotate32(
  Rotate32(
  Rotate32(h +
  Rotate32(Rotate32(g, 11) * c1, 17) * c1, 19) * 5+ $e6546b64, 17) * c1 +
  Rotate32(Rotate32(f, 11) * c1, 17) * c1, 19) * 5+ $e6546b64, 17) * c1;
end;

{$ENDIF}

end.
